/**********************************************************************
 *
 * LocARNA: LOCal Alignment of RNA
 *
 * Copyright (C) Sebastian Will <will(@)informatik.uni-freiburg.de> 
 *               2005, 2006
 *
 **********************************************************************/

/*

BUG:
Erzwingen der ungapped regions nicht korrekt (rechtes Ende und
im Zusammenhang mit arc matches)

*/


#include <iostream>
#include <fstream>
// #include <sstream>
#include <vector>

#include <queue>

#include <math.h>

#include "sequence.hh"
#include "basepairs.hh"
#include "alignment.hh"
#include "aligner.hh"


using namespace std;

const string VERSION_STRING="LocARNA 0.99 --- pairwise Local Alignment of RNAs";

// ------------------------------------------------------------
// Parameter

double min_prob; // only pairs with a probability of at least min_prob are taken into account

int match_score;
int mismatch_score;
int indel_score;

int structure_weight;

int tau_factor; // contribution of sequence similarity in an arc match (in percent)
               
bool no_lonely_pair; // no lonely pairs option

bool STRUCT_LOCAL; // allow exclusions for maximizing alignment of connected substructures 
bool SEQU_LOCAL; // maximize alignment of subsequences

const bool DO_TRACE=true;


int max_diff; // maximal difference for positions of alignment edges
              // (only used for ends of arcs)
int max_diff_am; //maximal difference between two arc ends, -1 is off

// only consider arc matchs where
//   1. for global (bl-al)>max_diff || (br-ar)<=max_diff    (if max_diff>=0)
//   2. for local (ar-al)-(br-bl)<=max_diff_am              (if max_diff_am>=0)
// except when there is no additional computation of M matrices necessary,
// this occurs if arcs are left-incident with larger arcs where 1 and 2 hold


int exclusion_score; // Score contribution per exclusion
                     // set to zero for unrestricted structure locality

double prob_exp_f(int seqlen) {return 1.0/(2*seqlen);} // expected probability of a base pair (null-model)

double exp_prob;
bool opt_exp_prob;

int output_width;

// ------------------------------------------------------------
// File arguments
char *bpsfile1;
char *bpsfile2;

char *clustal_out;
bool opt_clustal_out;

char *pp_out;
bool opt_pp_out;

// ------------------------------------------------------------
//
// Options
//
#include "options.hh"


int opt_kbest;
bool opt_help;
bool opt_version;
bool opt_verbose;
bool opt_local_output;
bool opt_pos_output;

bool opt_stacking;

char *ungapped_regionsA;


option_def my_options[] = {
    {"help",'h',&opt_help,O_NO_ARG,0,0,0,"This help"},
    {"version",'V',&opt_version,O_NO_ARG,0,0,0,"Version info"},
    {"verbose",'v',&opt_verbose,O_NO_ARG,0,0,0,"Verbose"},
    {"match",'m',0,O_ARG_INT,&match_score,"50","w","Match score"},
    {"mismatch",'M',0,O_ARG_INT,&mismatch_score,"0","w","Mismatch score"},
    {"indel",'i',0,O_ARG_INT,&indel_score,"-200","w","Indel score"},
    {"struct-weight",'s',0,O_ARG_INT,&structure_weight,"100","w","Maximal weight of 1/2 arc match"},
    {"tau",'t',0,O_ARG_INT,&tau_factor,"0","f","Tau factor in percent"},
    {"min-prob",'p',0,O_ARG_DOUBLE,&min_prob,"1e-2","p","Minimal probability"},
    {"exp-prob",'e',&opt_exp_prob,O_ARG_DOUBLE,&exp_prob,0,"p","Expected probability"},
    {"noLP",0,&no_lonely_pair,O_NO_ARG,0,0,0,"No lonely pair"},
    {"struct-local",0,0,O_ARG_BOOL,&STRUCT_LOCAL,"false","bool","Structure local"},
    {"sequ-local",0,0,O_ARG_BOOL,&SEQU_LOCAL,"true","bool","Sequence local"},
    {"clustal",0,&opt_clustal_out,O_ARG_STRING,&clustal_out,0,"file","Clustal output"},
    {"pp",0,&opt_pp_out,O_ARG_STRING,&pp_out,0,"file","PP output"},
    {"local-output",'L',&opt_local_output,O_NO_ARG,0,0,0,"Output only local sub-alignment"},
    {"pos-output",'P',&opt_pos_output,O_NO_ARG,0,0,0,"Output only local sub-alignment positions"},
    {"width",'w',0,O_ARG_INT,&output_width,"60","columns","Output width"},
    {"kbest",'k',0,O_ARG_INT,&opt_kbest,"-1","k","Find k-best alignments"},
    {"max-diff-match",'d',0,O_ARG_INT,&max_diff,"-1","diff","Maximal difference <d> for alignment edges"},
    {"max-diff-am",'D',0,O_ARG_INT,&max_diff_am,"-1","diff","Maximal difference <d> for sizes of matched arcs"},
    {"exclusion",'E',0,O_ARG_INT,&exclusion_score,"0","w","Exclusion weight"},
    {"ungappedA",'u',0,O_ARG_STRING,&ungapped_regionsA,"","range(s)","Ungapped region(s) in A"},
    {"stacking",0,&opt_stacking,O_NO_ARG,0,0,0,"Alignment with stacking terms"},
    {0,0,0,O_ARG_STRING,&bpsfile1,0,"bps-file 1","Basepairs input file 1"},
    {0,0,0,O_ARG_STRING,&bpsfile2,0,"bps-file 2","Basepairs input file 2"},
    {0,0,0,0,0,0,0,0}
};


// ------------------------------------------------------------


/* ------------------------------------------------------------
   Experimental feature: adapt score to sequence identity
   ------------------------------------------------------------ */

/*
prob_equals returns "probability" that two alignment columns are equal
this function needs to be re-implemented

double seq_identity(Sequence a, Sequence b) {
    const size_t n=a.length();
    const size_t m=b.length();

    vector<vector<double> > M;

    M.resize(n+1);
    for (uint i=0; i<=n; i++) M[i].resize(m+1);
        
    for (uint i=0; i<=n; i++) M[i][0]=0;
    for (uint j=0; j<=m; j++) M[0][j]=0;

    for (uint i=1; i<=n; i++)
	for (uint j=1; j<=m; j++)
	    M[i][j] = max(M[i-1][j-1] + a[i].prob_equals(b[j]),
			  max(M[i-1][j],
			      M[i][j-1]));
			      
    return 2.0*M[n][m]/(double)(n+m);
}
*/


/* ------------------------------------------------------------ */


/* ------------------------------------------------------------
   Compute k-best alignments by intervall splitting
*/

template <class T>
class greater_second {
public:
    bool operator() (const T& a, const T&b) {
	return a.second < b.second;
    }
};


typedef std::pair<AlignerRestriction,score_t> task_t;


//! write the k best local alignments
void kbest(Aligner &a, uint k) {

    // compute alignment score for a
    score_t a_score = a.align();
    
    // make priority queue tasks
    std::priority_queue<task_t, vector<task_t>, greater_second<task_t> > tasks;
    
    // put a into tasks
    tasks.push(task_t(a.get_restriction(),a_score));

    for (int i=1; i<=k; i++) {
	//std::cout << "queue size: "<<tasks.size()<<std::endl;
	
	// get best task from tasks
	task_t task = tasks.top();

	// pop topmost element
	tasks.pop();
	
	AlignerRestriction &task_r=task.first;
	score_t task_score=task.second;
	
	a.set_restriction(task_r);
	a.align(); // alignment needs to be recomputed (this is fast)!
	// do backtrace and write tasks alignment
	a.trace();
	Alignment alignment = a.get_alignment();
	
	// std::cout << "SCORE: " << task_score << std::endl;

	alignment.write(cout, output_width, task_score, 
			opt_local_output, opt_pos_output);
	
	if (!opt_pos_output) std::cout << std::endl
				       << std::endl;

	if (i==k) break; // break if enough solutions generated
	
	// split the longer sequence according to local alignment
	uint lenA=task_r.get_endA()-task_r.get_startA();
	uint lenB=task_r.get_endB()-task_r.get_startB();
	
	// make two clones of AlignerRestriction task_r
	AlignerRestriction r1(task_r);
	AlignerRestriction r2(task_r);
	
	if (lenA>lenB) {
	    // split A
	    int splitA = (alignment.get_local_startA() + alignment.get_local_endA())/2;
	    if (opt_verbose) std::cout <<"Split A at "<<splitA<<std::endl;
	    r1.set_endA(splitA);
	    r2.set_startA(splitA);
	} else {
	    int splitB = (alignment.get_local_startB() + alignment.get_local_endB())/2;
	    if (opt_verbose) std::cout <<"Split B at "<<splitB<<std::endl;
	    r1.set_endB(splitB);
	    r2.set_startB(splitB);
	}
	
	// compute alignment scores for both splits
	
	a.set_restriction(r1);
	score_t a1_score = a.align();
	a.set_restriction(r2);
	score_t a2_score = a.align();
	
	// std::cout <<"a1_score: " << a1_score << std::endl;
	// std::cout <<"a2_score: " << a2_score << std::endl;
	
	
	// put both splits into <tasks>
	
	tasks.push(task_t(r1,a1_score));
	tasks.push(task_t(r2,a2_score));
	
    }
}



void parseRangeList(const string &ungapped_regionsA, std::vector<bool> &ungappedA) {
    if (ungapped_regionsA.length()==0) return;
    string s=ungapped_regionsA+",";
    string val_str="";
    int val;
    int val_first=INT_MIN;
    for (int i=0; i!=s.length(); i++) {
	if (s[i]==',' || s[i]=='-') {
	    val=atoi(val_str.c_str());
	    val_str="";
	    if (s[i]=='-') {
		val_first=val;
	    } else {
		if (val_first==INT_MIN) {
		    val_first=val;
		}

		// std::cout << val_first << " - " << val << std::endl;
		if (val_first<1 || val>=ungappedA.size()) {
		    std::cerr <<"Parse error \"out of range\" in parseRangeList" << std::endl;
		    exit(-1);
		}
		for (int j=val_first; j<=val; j++) ungappedA[j]=true;
		val_first=INT_MIN;
	    }
	} else {
	    val_str+=s[i];
	}
    }
}


// ------------------------------------------------------------





// ------------------------------------------------------------
// MAIN

int
main(int argc, char **argv) {
    
    bool process_success=process_options(argc,argv,my_options);

    if (opt_help) {
	cout << VERSION_STRING<<endl<<endl;
	print_help(argv[0],my_options);
	exit(0);
    }
    
    if (opt_version) {
	cout << VERSION_STRING<<endl<<endl;
	exit(0);
    }

    if (!process_success) {
	puts(O_error_msg);
	printf("USAGE: ");
	print_usage(argv[0],my_options);
	printf("\n");
	exit(-1);
    }
        
    if (opt_verbose)
	print_options(my_options);
    
    Sequence seqA;
    Sequence seqB;


    BasePairsParams bpp(structure_weight, min_prob, 
			exp_prob, opt_exp_prob);

    BasePairs bpsA(&bpp);
    BasePairs bpsB(&bpp);
    
    bpsA.read(bpsfile1,seqA,opt_stacking);
    bpsB.read(bpsfile2,seqB,opt_stacking);
    
    if (opt_verbose) {
	cout << "Sequence A: "<<endl;
	seqA.write(cout);
	cout<<" (Length:"<< seqA.length()<<", Basepairs:"<<bpsA.num_bps() << ")" <<endl;
		
	cout << "Sequence B: "<<endl;
	seqB.write(cout);
	cout<<" (Length:"<< seqB.length()<<", Basepairs:"<<bpsB.num_bps() << ")" <<endl;

	// cout << "Base Identity: "<<(seq_identity(seqA,seqB)*100)<<endl; 
    }


    // ----------------------------------------
    // parse ungapped regions of A
    
    std::vector<bool> ungappedA(seqA.length()+1);

    parseRangeList(ungapped_regionsA,ungappedA);
    
    if (opt_verbose && strlen(ungapped_regionsA)>0) {
	for (int i=1; i<ungappedA.size(); i++) std::cout << ungappedA[i];
	std::cout << std::endl;
    }

    // ----------------------------------------

    
    AlignerParams aligner_params(match_score, mismatch_score, indel_score,
				 tau_factor, no_lonely_pair,
				 STRUCT_LOCAL, SEQU_LOCAL,
				 exclusion_score, max_diff, max_diff_am,
				 ungappedA);

    Aligner aligner(seqA,bpsA,seqB,bpsB,&aligner_params);

    // if option --kbest is given, use kbest algo (by interval splitting)
    if (opt_kbest>0) {
	kbest(aligner,opt_kbest);
	exit(0);
    }


    // otherwise find the best alignment
	
    score_t score = aligner.align();
    
    cout << "Score: "<<score<<endl;
    
    if (DO_TRACE) {
	 aligner.trace();
	 aligner.get_alignment().write(cout, output_width, score,
				       opt_local_output, opt_pos_output);
	 cout<<endl;

	 if (opt_clustal_out) {
	     ofstream out(clustal_out);
	     if (out) {
		 aligner.get_alignment().write_clustal(out, output_width,
						       score,
						       opt_local_output,
						       opt_pos_output);
	    } else {
		cerr << "Cannot write to "<<clustal_out<<endl<<"! Exit.";
		exit(-1);
	    }
	}
	if (opt_pp_out) {
	    ofstream out(pp_out);
	    if (out) {
		aligner.get_alignment().
		    write_pp(out,bpsA,bpsB,output_width);
	    } else {
		cerr << "Cannot write to "<<pp_out<<endl<<"! Exit.";
		exit(-1);
	    }
	}
    }
    exit(0);
}
